package Fork;

use strict;
use Carp;
use Exporter;
use Data::Dumper;
use Socket;
use IO::Handle;
use IO::Select;

use vars qw($VERSION @ISA @EXPORT );
@ISA = qw/Exporter/;
@EXPORT = qw/fork_it/;
$VERSION='1.0.0';

sub fork_it($) {
    my $param = shift;

    # How many children to spawn ?
    my $max_spawned = scalar @{$param->{args}} <
	$param->{max_children} ? scalar @{$param->{args}} :
	$param->{max_children};

    # Do we need a timeout ?
    $param->{timeout} ||= 0;

    # Spawn the children
    my ($read_select, $write_select)=(new IO::Select, new IO::Select);
    local $/ = "\n\n";
    my $VAR1;
    foreach (1..$max_spawned) {
	my($child_fh, $parent_fh) = (new IO::Handle, new IO::Handle);
	socketpair($child_fh, $parent_fh, AF_UNIX, SOCK_STREAM,
		   PF_UNSPEC) || die "socketpair failed: $!\n";
	$child_fh->autoflush;
	$parent_fh->autoflush;

	if (my $pid = fork) {
	    close $parent_fh;
	    # Start out writing to all children
	    $write_select->add( $child_fh );
	} else {
	    # Child starts here
	    croak "cannot fork: $!\n" unless defined $pid;

	    # close unwanted file handles
	    close $child_fh;
	    close STDIN if !defined $param->{keep_stdin};
	    close STDOUT if !defined $param->{keep_stdout};

	    local $SIG{'ALRM'} = sub { die "timeout\n" };
	    my $ret;
	    while(<$parent_fh>) {
		# get data
		eval "$_";
		$ret = ();
 		eval {
 		    alarm $param->{timeout};
 		    $ret = $param->{fct}->($VAR1);
 		    alarm 0;
 		};
 		# send returned value
		$ret = "E: timeout" if defined $@ && $@ eq "timeout\n";
 		print $parent_fh Dumper($ret), "\n";
	    }
	    # kill the child
	    exit 0;
	}
    }

    my $ret = ();
    my $num_data=0;
    my $read_fh;
    my $write_fh;
    my $data = '';

    while(1) {
	my($readable, $writable) =
	    IO::Select->select($read_select, $write_select, undef);
	# Send data to children
	if (@$writable) {
	    while(@$writable &&
		  defined($data = shift @{$param->{args}})) {
		# Send data to child
		$write_fh = shift @$writable;
		print $write_fh Dumper($data), "\n";
		# Move to read set
		$write_select->remove($write_fh);
		$read_select->add($write_fh);
		$num_data++;
	    }
	    # Are we done with input?
	    defined $data || undef $write_select;
	}

	# One or more children have an answer
	while (@$readable) {
	    # read data from children
	    $read_fh = shift @$readable;
	    eval <$read_fh>;
	    push @$ret, $VAR1;
	    # Move to write set
	    $read_select->remove($read_fh);
	    $write_select->add($read_fh) if defined $write_select;
	    $num_data--;
	}
	last if !defined $write_select && !$num_data;
    }
    return $ret;
}


1;
__END__

=pod

=head1 NAME

Fork - A fork(2) wraper for Perl

=head1 SYNOPSIS

use Fork;

    my $values = fork_it \%arguments;

=head1 DESCRIPTION

B<Fork> makes fork development easier.

=head1 Function B<for_it>

=head2 Syntax

    fork_it(\%arguments);

=head2 Arguments

Valid keys un the I<argument> hash reference are:

=over

=item B<fct>: is a reference to a one-parameter function which should
return either undef or a reference.

=item B<args>: is a reference to the parameters for I<fct>.

=item B<max_children>: maximun children to launch

=item B<timeout>: maximun time (in seconds) allowed for each I<fct>
call.

=item B<keep_stdin>: Do not close child's STDIN.

=item B<keep_stdout>: Do not close child's STDOUT.


=back

=head2 Return value

B<fork_it> returns an array reference which contains return values of
all children. Logically the array I<should NOT be sorted>. Thus you
have to sort data yourself. See B<EXAMPLES> for futher information.

=head1 EXAMPLES

See B<examples> directory.

=head1 AUTHOR

Sebastien J. Gross <seb@sjgross.org>
All rights reserved. This program is free software; you can redistribute
it and/or modify it under the terms of the GPL.

=head1 VERSION

This code is beta, version 1.0.0
Copyright (c) 2002 Sebastien J. Gross. All rights reserved.
This program is free software; you can redistribute it and/or modify
it under the terms of the GPL.

=cut
